@@ -1,7 +1,7 @@
 /*
  * Part of DNS zone file validator `validns`.
  *
- * Copyright 2011, 2012 Anton Berezin <tobez@tobez.org>
+ * Copyright 2011-2013 Anton Berezin <tobez@tobez.org>
  * Modified BSD license.
  * (See LICENSE file in the distribution.)
  *
@@ -17,6 +17,7 @@
 #include "carp.h"
 #include "textparse.h"
 #include "rr.h"
+#include "cbtree.h"
 
 static char* rdtype2str_map[T_MAX+1] = {
 	"0",
@@ -85,7 +86,7 @@ static char* rdtype2str_map[T_MAX+1] = {
 	"L64",
 	"LP",
 };
-void *zone_data = NULL;
+struct cbtree zone_data = {NULL};
 char *zone_apex = NULL;
 int zone_apex_l = 0;
 
@@ -102,6 +103,9 @@ char *rdtype2str(int type)
 	}
 	r = rdtype2str_map[type];
 	if (r)	return r;
+	if (type == 32769) {
+		return rdtype2str_map[type] = "DLV";
+	}
 	sprintf(s, "TYPE%d", type);
 	return quickstrdup_temp(s);
 }
@@ -153,6 +157,8 @@ struct binary_data name2wire_name(char *s)
 	*c = (unsigned char)(r-c-1);
 	toret.length = r-res;
 	toret.data = (char*)res;
+	if (toret.length == 2) /* "." is just 00, not 00 00 */
+	    toret.length = 1;
 	return toret;
 }
 
@@ -172,15 +178,15 @@ static struct named_rr *find_or_create_named_rr(char *name)
 		named_rr->flags = 0;
 		named_rr->parent = NULL;
 
-		JSLI(named_rr_slot, zone_data, name2findable_name(name));
-		if (named_rr_slot == PJERR)
-			croak(2, "find_or_create_named_rr: JSLI failed");
+		named_rr_slot = (void *)cbtree_insert(&zone_data, (char *)name2findable_name(name));
+		if (!named_rr_slot)
+			croak(2, "find_or_create_named_rr: tree insertion failed");
 		if (*named_rr_slot)
 			croak(3, "find_or_create_named_rr: assertion error, %s should not be there", name);
 		*named_rr_slot = named_rr;
 		G.stats.names_count++;
 
-		s = index(name, '.');
+		s = strchr(name, '.');
 		if (s && s[1] != '\0') {
 			named_rr->parent = find_or_create_named_rr(s+1);
 		}
@@ -363,7 +369,7 @@ struct named_rr *find_named_rr(char *name)
 {
 	struct named_rr **named_rr_slot;
 
-	JSLG(named_rr_slot, zone_data, name2findable_name(name));
+	named_rr_slot = (void*) cbtree_find(&zone_data, (char *)name2findable_name(name));
 	if (named_rr_slot)
 		return *named_rr_slot;
 	return NULL;
@@ -371,14 +377,12 @@ struct named_rr *find_named_rr(char *name)
 
 struct named_rr *find_next_named_rr(struct named_rr *named_rr)
 {
-	unsigned char sorted_name[512];
-	struct named_rr **named_rr_p;
+	struct named_rr *res;
 
-	strcpy((char*)sorted_name, (char*)name2findable_name(named_rr->name));
-	JSLN(named_rr_p, zone_data, sorted_name);
-	if (named_rr_p)
-		return *named_rr_p;
-	return NULL;
+	if (cbtree_next(&zone_data, (char *)name2findable_name(named_rr->name), (intptr_t *)&res) == NULL)
+		return NULL;
+
+	return res;
 }
 
 struct rr_set *find_rr_set(int rdtype, char *name)
@@ -469,6 +473,8 @@ int str2rdtype(char *rdtype, int *is_generic)
 			return T_A;
 		} else if (strcmp(rdtype, "aaaa") == 0) {
 			return T_AAAA;
+		} else if (strcmp(rdtype, "afsdb") == 0) {
+			return T_AFSDB;
 		}
 		break;
 	case 'c':
@@ -485,6 +491,10 @@ int str2rdtype(char *rdtype, int *is_generic)
 			return T_DNSKEY;
 		} else if (strcmp(rdtype, "dname") == 0) {
 			return T_DNAME;
+		} else if (strcmp(rdtype, "dlv") == 0) {
+			return T_DLV;
+		} else if (strcmp(rdtype, "dhcid") == 0) {
+			return T_DHCID;
 		}
 		break;
 	case 'h':
@@ -495,6 +505,13 @@ int str2rdtype(char *rdtype, int *is_generic)
 	case 'i':
 		if (strcmp(rdtype, "ipseckey") == 0) {
 			return T_IPSECKEY;
+		} else if (strcmp(rdtype, "isdn") == 0) {
+			return T_ISDN;
+		}
+		break;
+	case 'k':
+		if (strcmp(rdtype, "kx") == 0) {
+			return T_KX;
 		}
 		break;
 	case 'l':
@@ -511,6 +528,14 @@ int str2rdtype(char *rdtype, int *is_generic)
 	case 'm':
 		if (strcmp(rdtype, "mx") == 0) {
 			return T_MX;
+		} else if (strcmp(rdtype, "mb") == 0) {
+			return T_MB;
+		} else if (strcmp(rdtype, "mg") == 0) {
+			return T_MG;
+		} else if (strcmp(rdtype, "minfo") == 0) {
+			return T_MINFO;
+		} else if (strcmp(rdtype, "mr") == 0) {
+			return T_MR;
 		}
 		break;
 	case 'n':
@@ -526,11 +551,15 @@ int str2rdtype(char *rdtype, int *is_generic)
 			return T_NID;
 		} else if (strcmp(rdtype, "nsec3param") == 0) {
 			return T_NSEC3PARAM;
+		} else if (strcmp(rdtype, "nsap") == 0) {
+			return T_NSAP;
 		}
 		break;
 	case 'p':
 		if (strcmp(rdtype, "ptr") == 0) {
 			return T_PTR;
+		} else if (strcmp(rdtype, "px") == 0) {
+			return T_PX;
 		}
 		break;
 	case 'r':
@@ -538,6 +567,8 @@ int str2rdtype(char *rdtype, int *is_generic)
 			return T_RRSIG;
 		} else if (strcmp(rdtype, "rp") == 0) {
 			return T_RP;
+		} else if (strcmp(rdtype, "rt") == 0) {
+			return T_RT;
 		}
 		break;
 	case 's':
@@ -564,6 +595,11 @@ int str2rdtype(char *rdtype, int *is_generic)
 			return type;
 		}
 		break;
+	case 'x':
+		if (strcmp(rdtype, "x25") == 0) {
+			return T_X25;
+		}
+		break;
 	}
 	bitch("invalid or unsupported rdtype %s", rdtype);
 	return -1;
@@ -619,8 +655,10 @@ void debug(struct named_rr *named_rr, char *s)
 	fprintf(stderr, "\n");
 }
 
-void validate_named_rr(struct named_rr *named_rr)
+static int
+validate_named_rr(const char *name, intptr_t *data, void *p)
 {
+	struct named_rr *named_rr = *((struct named_rr **)data);
 	Word_t rdtype;
 	struct rr_set **rr_set_p;
 	int nsec3_present = 0;
@@ -687,7 +725,7 @@ void validate_named_rr(struct named_rr *named_rr)
 	if (nsec3_present && nsec3_only) {
 		named_rr->flags |= NAME_FLAG_NSEC3_ONLY;
 	}
-//debug(named_rr, "<<<<");
+	return 1;
 }
 
 
@@ -733,38 +771,27 @@ static void* nsec_validate_pass2(struct rr *rrv)
 	return rr;
 }
 
-void second_validation_pass()
+static int
+second_pass_one_name(const char *name, intptr_t *data, void *p)
 {
-	unsigned char sorted_name[512];
-	struct named_rr **named_rr_p;
+	struct named_rr *named_rr = *((struct named_rr **)data);
+	struct rr_set **rr_set_p;
 
-	sorted_name[0] = 0;
-	JSLF(named_rr_p, zone_data, sorted_name);
-	while (named_rr_p) {
-		struct rr_set **rr_set_p;
-
-		freeall_temp();
-		JLG(rr_set_p, (*named_rr_p)->rr_sets, T_NSEC);
-		if (rr_set_p && (*rr_set_p)->tail) {
-			nsec_validate_pass2((*rr_set_p)->tail);
-		} else {
-		}
-		JSLN(named_rr_p, zone_data, sorted_name);
+	freeall_temp();
+	JLG(rr_set_p, named_rr->rr_sets, T_NSEC);
+	if (rr_set_p && (*rr_set_p)->tail) {
+		nsec_validate_pass2((*rr_set_p)->tail);
 	}
+	return 1;
 }
 
 void validate_zone(void)
 {
-	unsigned char sorted_name[512];
-	struct named_rr **named_rr_p;
+	cbtree_allprefixed(&zone_data, "", validate_named_rr, NULL);
+	cbtree_allprefixed(&zone_data, "", second_pass_one_name, NULL);
 
-	sorted_name[0] = 0;
-	JSLF(named_rr_p, zone_data, sorted_name);
-	while (named_rr_p) {
-		validate_named_rr(*named_rr_p);
-		JSLN(named_rr_p, zone_data, sorted_name);
-	}
-	second_validation_pass();
+	if (G.dnssec_active && !G.nsec3_present)
+		validate_nsec_chain();
 }
 
 void validate_record(struct rr *rr)
